#ifndef __GLOBALBACKPRESSURE_H__
#define __GLOBALBACKPRESSURE_H__

#include <packetforwarding.h>
#include <dlhandler.h>
#include <buffer.h>

class BufferQueue;
//! Schaltmatrix-Manipulator fuer globales Backpressure.
/*!
 * Diese Klasse sorgt fuer die Anpassung der Schaltmatrix eines Routers bezueglich 
 * der Bedingungen fuer das globale Backpressure. Dazu wird das Bit IO_FIT eines 
 * Einganges gesetzt, wenn der Netzknoten am zugehoerigen Ausgang ein zu vermittelndes 
 * Paket des zugehoerigen Einganges aufnehmen kann.
 *
 * @see WireMatrixEntry
 * 
 * @author Matthias Kuehm <kuehmi@cs.tu-berlin.de>
 */
class GBForwardingModifier : public ForwardingModifier
{
public:
	
	//! Re-Implemenierung der abstrakten Basisklasse WireMatrixModifier.
	/*!	 
	 * Damit ueberhaupt geprueft werden kann, ob der nachfolgende Puffer 
	 * ein Paket aufnehmen kann, muss ein Paket vorliegen, dass den den 
	 * Ausgang zum uebergebenen Schaltmatrixeintrag benutzen will. Kann
	 * dieser Ausgang noch senden, so kann noch keine Entscheidung getroffen
	 * werden.
	 * 
	 * return PASSED, falls Bits IO_VALID und I_PACKET gesetzt, FAILED 
	 * 	sonst. SKIPPED, falls Ausgang noch unklar.
	 */
	int precondition(WireMatrixIterator*) const;
};


//! Koroutine fuer globales Backpressure.
/*!
 * Diese Klasse implementiert eine Koroutine der Simulation, die fuer die 
 * Bewegung der Pakete im Netz waehrend eines Taktes nach der Strategie des 
 * globalen Backpressure sorgt. Dazu werden die Fuellstaende der Puffer im 
 * Netz betrachtet, wie sie durch Paketbewegungen im aktuellen Taktes vorliegen.
 *
 * @author Matthias Kuehm <kuehmi@cs.tu-berlin.de>
 */
class GBForwarding : public PacketForwarding
{
public:
	
	//! Konstruktor.
	/*!
	 * Der Konstruktor bekommt als Argument einen Ausschnitt des
	 * eingelesenen XML-Dokuments in Form eines Element-Knotens
	 * des vom Parser erstellten DOM-Baumes. Dieser Knoten (oder
	 * untergeordnete Knoten) liefert Informationen zu Routing,
	 * Switching und Backpressure-Mechanismus.
	 * 
	 * @param scope XML-Element mit Spezifikation der Koroutine.
	 *
	 * Eine Variante der Markierung von Puffern wurde von Christian
	 * Zimmermann vorgeschlagen und getestet, die aus 6 nderungen
	 * des Code in den Dateien gbforwarding.h, gbforwarding.cpp,
	 * buffer.cpp und router.cpp besteht. Sie wurde hier nur in 
	 * Kommentaren vermerkt, doch wird momentan nicht benutzt. 
	 *
	 * Herkmmlich: receivers werden PULLED signiert. D.H sie wer-
	 * den automatisch als gesendet behandelt, obwohl dies nicht
	 * unbedingt der Fall ist. (Wieso es kombiniert mit dlhandler
	 * trotzdem funktioniert, ist etwas schleierhaft.)
	 *
	 * Vorschlag: receivers als PUSHED signieren. Sinn: Sie sind
	 * zum empfangen vorgemerkt und drfen es nicht ein zweites Mal
	 * tun. PULLED signiert werden Puffer, die tatschlich gesendet
	 * haben. Nur diese drfen nicht mehr senden.
	 *
	 * Der Vorschlag funktioniert und liefert die gleichen Resul-
	 * tate mit gleicher CPU-Laufzeit, doch keine Verbesserung.
	 */
	GBForwarding(const DOMElement* scope=NULL) 
		// Vorschlag nderung 1: Receivers sollen PUSHED statt PULLED vermerkt werden.
		: PacketForwarding(scope), receivers(TriggerNode::PULLED/*PUSHED*/) {}
			
	//! Implementierung der abstrakten Basisklasse PacketForwarding.
	/*!
	 * Zur Realisierung von globalem Backpressure wird das Netz
	 * beginnend bei den Zielen und den leeren Puffern zu Beginn
	 * eines Taktes durchlaufen. Pakete duerfen nur zu Knoten (Puffern) 
	 * vermittelt werden, deren Fuellstand fuer diesen Takt geklaert ist, 
	 * die also gesendet haben oder in diesem Takt nicht mehr senden 
	 * werden. Diese Strategie kann bei zyklischen Routing-Verfahren zu 
	 * Deadlocks fuehren.
	 *
	 * @see PacketForwarding
	 */
	void start();
	
	
protected:
		
	//! Implementierung der abstrakten Basisklasse PacketForwarding.
	/*!
	 * Fuer globales Backpressure wird das Netz baumartig von den Zielen zu
	 * den Quellen durchlaufen. Es werden die Baumstruktur vorbereitet und
	 * eine Instanz der Klasse GBForwardingModifier geliefert.
	 *
	 * @see GBForwardingModifier
	 * @see PacketForwarding
	 * @see initActivationTress
	 * 
	 * @return Instanz der Klasse GBForwardingModifier.
	 */
	ForwardingModifier* initForwarding();
	
	//! Aktivieren eines Knotens.
	/*!
	 * Das Aktivieren eines Knotens sorgt dafuer, dass geprueft wird, ob
	 * ein Paket in diesem Takt zu versenden ist. Ist dies der Fall, so
	 * wird dieses entsprechend an eine ausgehende Leitung angelegt. Bei
	 * einem Router wird daraufhin die Schaltmatrix mit entsprechenden
	 * Routing-Informationen aktualisiert.
	 *
	 * @param tnode Ein zu aktivierender Knoten.
	 * @return Status (e_trigger_return).
	 */
	int activate(TriggerNode* tnode);
	
	//! Anfordern eines Paketes.
	/*!
	 * Bevor ein Paket empfangen werden kann, muss ein Knoten seine
	 * Bereitschaft seinem Vorgaenger signalisieren. Das sorgt bei einem
	 * Router dazu, dass um den entsprechenden Ausgang bei Bedarf ein
	 * Scheduling der Eingaenge durchgefuehrt wird.
	 *
	 * @param tnode Ein empfangsbereiter Knoten.
	 * @return Status (e_trigger_return).
	 */
	int request(TriggerNode* tnode);
	
	//! Aktivieren der Kind-Knoten.
	/*!
	 * Bevor ein Knoten seine Empfangsbereitschaft signalisiert, werden seine 
	 * Vorgaenger-Knoten aktiviert, wodurch sie ihre zu versendenden Pakete an 
	 * ihre Ausgaenge anlegen. Leere oder inaktive Knoten, werden zur Liste
	 * empfangsbereiter knoten hinzugefuegt.
	 * 
	 * @param parent Ein TriggerNode mit bekannter Aufnahmekapazitaet.
	 * @param it Einfuegeposition fuer leere oder inaktive Knoten in
	 * Liste empfangsbereiter Knoten.
	 */
	void activateChildren(TriggerNode* parent, TNodeList::iterator it);

	//! Empfangen der Pakete.
	/*!
	 * Das Empfangen von Paketen sorgt fuer die Paketbewegung im Netz.
	 * Dazu wird eine Liste aller empfangsbereiten Knoten durchlaufen
	 * und modifiziert. Ein empfangsbereiter Knoten wird aus der Liste
	 * entfernt (vorn), seine Vorgaenger aktiviert und ein Paket 
	 * angefordert. Kann ein Paket empfangen werden, so wird in Richtung 
	 * der nun gesendeten oder nicht mehr sendenden Knoten weitergelaufen.
	 * Kann kein Paket empfangen werden, so wird im Falle eines vollen
	 * Puffers in Richtung der nicht mehr vermittelbaren Vorgaenger
	 * weiter traversiert. Leere oder blockierte Knoten werden bereits 
	 * bei der Aktivierung als gesendet markiert und zur Liste der 
	 * empfangsbereiten Knoten hinzugefuegt (vorn). Bei Blockierungen 
	 * durch einen Router wird ein Knoten wieder zur Liste hinzugefuegt 
	 * (hinten). Kommt er erneut an die Reihe und kann vom Router noch 
	 * immer keine Entscheidung getroffen werden, so wird abgebrochen.
	 */
	void receivePackets();

	//! Initialisierung der baumartigen Vernetzung der TriggerNodes.
	/*!
	 * Diese Methode initialisiert die zu triggernden Netz-Knoten.
	 * Dazu werden beginnend bei den Zielen, die Vorgaenger eines
	 * jeden Netzknotens erfasst, wobei nicht zu triggernde Knoten,
	 * wie Router, uebersprungen werden.
	 */
	void initActivationTrees();
	
	//! Nachfolgersuche.
	/*!
	 * Bei Umwegpaketen muss das GBForwarding einen neuen Bezugspuffer
	 * finden, der kein Target ist.
	 */
	vector<Buffer*> spaceSuccessors(Buffer* source);
	
private:
	
	//! Initiale Liste empfangsbereiter Knoten (Targets und leere Puffer).
	TNodeVector initial;
	//! Liste empfangsbereiter Knoten durch Netztraversierung.
	SignedTNodeList receivers;
	//! Ist der Puffer k schon auf Nachfolger geprft worden?
	vector<bool> checked;
};

#endif

