#include <lbforwarding.h>
#include <configuration.h>
#include <sourcebuffer.h>
#include <bufferqueue.h>
#include <macros.h>
#include <utility/xmlhelper.h>
#include <utility/random.h>
#include <routing.h>
#include <switching.h>
#include <wirematrixiterator.h>
#include <flit.h>

#include <utility/bitmask.h>

#include <cassert>

REGISTER_ELEMENT(LBForwarding, "LocalBackpressure")


void LBForwarding::start()
{
	receivers.clear();
	
	const ComponentVector& sources = Components(Component::SOURCE);
	const ComponentVector& targets = Components(Component::TARGET);
	const ComponentVector& buffers = Components(Component::BUFFER);

	unsigned int bcount = buffers.size();

	// alle Generatoren aktivieren
	for(unsigned int i=0; i<sources.size(); i++)
	{
		activate((TriggerNode*)sources[i]);
	}

	// alle Puffer aktivieren
	for(unsigned int i=0; i<bcount; i++)
	{
		activate((TriggerNode*)buffers[i]);
	}

	// alle Puffer und Ziele fordern in zufaelliger Reihenfolge ein Flit...
	RandomKofN b(bcount+targets.size());
	while(!b.Empty())
	{	
		unsigned int chosen = b.Choose()-1;
		if(chosen >= bcount)
			request((TriggerNode*)targets[chosen-bcount]);
		else
			request((TriggerNode*)buffers[chosen]);
	}
	
	// Pakete bewegen
	receivePackets();

#ifdef PACKETMOVE
	
	for(unsigned int i=0; i<buffers.size();i++)
	{
		if(!((TriggerNode*)buffers[i])->test(TriggerNode::ACTIVE))
			_ERROR(buffers[i]->toString() << " not activated in this tick!", 111);

		if(((TriggerNode*)buffers[i])->In().size() 
				&& !((TriggerNode*)buffers[i])->test(TriggerNode::PUSHED))
			_ERROR(buffers[i]->toString() << " not received in this tick!", 111);
	}

	for(unsigned int i=0; i<targets.size(); i++)
	{
		if(((TriggerNode*)targets[i])->In().size() 
				&& !((TriggerNode*)targets[i])->test(TriggerNode::PUSHED))
			_ERROR(targets[i]->toString() << " not received in this tick!", 111);
	}
	
	for(unsigned int i=0; i<sources.size();i++)
	{
		if(!((TriggerNode*)sources[i])->test(TriggerNode::ACTIVE))
			_ERROR(sources[i]->toString() << " not activated in this tick!", 111);
	}
	
#endif

}

void LBForwarding::receivePackets()
{
	// empfangsbereiter Knoten
	TriggerNode* tnode;
	// eingehende Leitung
	Wire* inwire;
	// Paketende erreicht
	bool tailOnLine = false;
	// Modifier zum Freigeben einer Schaltung
	ModifierList clear;
	clear.push_back(switching()->clear());
	
	// Solange empfangsbereite Knoten existieren
	while(!receivers.empty())
	{		
		// erster Knoten aus Liste der moeglich Empfaenger
		tnode = receivers.front();
		receivers.pop_front();
		
		// Eingangsleitung
		inwire = tnode->In().front();

		TriggerNode* prev = inwire->getFlit(tnode)->Location().where;

		// Flit empfangen
		unsigned int status = tnode->takePacketFromLine(inwire, tailOnLine);
			
		// ggf. Schaltung aufheben
		if(status==TriggerNode::SUCCESS && tailOnLine)
		{
			// Letztes Flit eines Paketes -> Schaltung aufheben
			inwire->From()->lineFinished(inwire, &clear);

			//Buffer und sein Vorgnger zum Beschneiden vormerken.
			if(tnode->rtti()==1 && ((Buffer*)tnode)->BQ->infinite)
				((Buffer*)tnode)->BQ->cut.sign();
			if(prev->rtti()==1 && ((Buffer*)prev)->BQ->infinite)
				((Buffer*)prev)->BQ->cut.sign();
			// Maeander-Puffer falls gesendet zum auflsen vormerken.
			if(prev->rtti()==1 && ((Buffer*)prev)->maeander)
				((Buffer*)prev)->solve_maeander.sign();
		}
		else
			inwire->From()->lineFinished(inwire, NULL);
		
		// Knoten markieren
		tnode->set(TriggerNode::PUSHED);
	}
}

int LBForwarding::activate(TriggerNode* tnode)
{
	ModifierList matrixinit;
	bool headOnLine = false;	
	
	// Nur Targets haben keine Nachfolger und knnen normalerweise
	// auch nicht senden...
	if(tnode->Out().size() != 1)
		return TriggerNode::DISMISS;

	// Der nachfolgende Netzknoten
	Node* next = tnode->Out().front()->To(); 
	
	// Paket auf die Line
	int state = tnode->putPacketOnLine(tnode->Out().front(), headOnLine);
	switch(state)
	{
		case TriggerNode::SUCCESS:
		{
			if(headOnLine)
				// Head liegt auf der Leitung, Routing-Methode
				// bei Benachrichtigung des Nachfolgers uebergeben
				matrixinit.push_back(routing());
		}
		case TriggerNode::NOPACKET:break;
		default:
			_ERROR("Unexpected state after activating line!", 111);
	}
	
	// Nachfolger benachrichtigen, ggf. Routing durchfuehren
	next->lineActivated(tnode->Out().front(), &matrixinit);

	// Knoten als aktiviert markieren
	tnode->set(TriggerNode::ACTIVE);
	
	return state;
}

int LBForwarding::request(TriggerNode* tnode)
{	
	// keine eingende Leitung, kein Paket  
	if(!tnode->In().size())
	{
		tnode->set(TriggerNode::PUSHED);
		return TriggerNode::NOPACKET;
	}
	
	// eingehende Leitung mit eventl. anliegendem Flit
	Wire* inwire = tnode->In().front();
	Node* prev = inwire->From();
	bool headOnLine = false;
	
	// kein Register auf Leitung, d.h. keine Schaltung
	if(!inwire->RegisterOnLine())
	{
		// Annahme: Vorgaengerknoten ist ein Router
		assert(prev->rtti() == Component::ROUTER);
		
		// angeschlossenen Router aktivieren
		switch(prev->putPacketOnLine(inwire, headOnLine))
		{
			// Kein Paket oder verloren
			case TriggerNode::NOPACKET:
			{
				// Annahme: Leitung frei
				assert(!inwire->RegisterOnLine());
						
				prev->lineFinished(inwire, NULL);
				tnode->set(TriggerNode::PUSHED);
				return TriggerNode::NOPACKET;
			}

			// Neue Schaltung hergestellt; Nachfolgender Knoten kann
			// ganzes Paket oder Flit aufnehmen; Flit bereit
			case TriggerNode::SUCCESS:
			{
				// Annahme: Flit fuer diesen Knoten verfuegbar
				assert(inwire->getFlit(tnode) != NULL);
				// Annahme: Head-Flit
				assert(headOnLine);
				// Annahme: Flit darf bewegt werden
				assert(inwire->getFlit(tnode)->moveable());
				
				// Knoten nun ist empfangsbereit		
				receivers.push_back(tnode);
				return TriggerNode::SUCCESS;
			}
			default:
				_ERROR("Unexpected status after request for packet!", 111);
		}
	}

	const Flit* fptr = inwire->getFlit(tnode);
	
	// Schaltung vorhanden; Pruefen, ob Flit verfuegbar
	if(!fptr ||	
		// Pruefen, ob Pufferstand vom Ende des letzten Taktes fuer Flit/Paket ausreicht
		!((Buffer*)tnode)->BQ->freeSpace() 
			|| (((Buffer*)tnode)->BQ->estimatedSpace() < fptr->neededSpace()))
	{
		// Signal, dass Leitung fertig
		prev->lineFinished(inwire, NULL);
		
		// Knoten markieren
		tnode->set(TriggerNode::PUSHED);
		return TriggerNode::DISMISS;
	}
		
	// Flit vorhanden und Knoten aufnahmebereit
	receivers.push_back(tnode);
	return TriggerNode::SUCCESS;
}

int LBForwardingModifier::precondition(WireMatrixIterator* it) const
{
	WireMatrixEntry& entry = it->value();

	if(entry.getOutWire()->To()->rtti() == Component::ROUTER)
		_ERROR("Can not handle a router at outgoing wire!", 111);

	return entry.test(I_PACKET) && entry.test(IO_VALID);
}

