#include <environment/analysis/test_estimation.h>
#include <environment/analysis/independentmean.h>
#include <environment/analysis/spectralmean.h>
#include <environment/analysis/spectralquantile.h>

#include <environment/prng.h>
#include <macros.h>

#include <iostream>
#include <cmath>
#include <cfloat>

#define _RUNS_ 1000
#define _p_ 0.25
#define _conf_ ((double)0.95)
#define _prec_ ((double)0.05)

CPPUNIT_TEST_SUITE_REGISTRATION( Test_IndependentMean );
CPPUNIT_TEST_SUITE_REGISTRATION( Test_SpectralMean );
CPPUNIT_TEST_SUITE_REGISTRATION( Test_SpectralQuantile );

struct TestDistribution
{
	virtual double nextValue() const = 0;
	
	double mean;

	double quantile;
};


Estimation* Test_IndependentMean::create()
{
	return new IndependentMean(_conf_);
}

Estimation* Test_SpectralMean::create()
{
	return new SpectralMean(_conf_);
}

Estimation* Test_SpectralQuantile::create()
{
	return new SpectralQuantile(_conf_, _p_);
}

void Test_Estimation::setUp()
{
	_est = create();
	_fmax = (unsigned int)_RUNS_ - (unsigned int)rint((double)_RUNS_ * _conf_);
	
	//std::cerr << _fmax << std::endl;
	
	_error = DBL_MAX;	
}

void Test_Estimation::tearDown()
{
	delete _est;
}

void Test_Estimation::uniform_check()
{
	struct : public TestDistribution
	{
		double nextValue() const
		{
			return UniformDouble();
		}
		
	} uniform;

	uniform.mean = 0.5;
	uniform.quantile = _p_;

	applyDistribution(&uniform);	
}


void Test_Estimation::exponential_check()
{
	double lambda = 0.5;
	
	struct : public TestDistribution
	{
		double nextValue() const
		{
			return -1.0 * mean * log(UniformDouble());
		}
		
	} exponential;

	exponential.mean = 1.0/lambda;
	exponential.quantile = -1.0 * 1.0/lambda * log(1.0 - _p_);

	applyDistribution(&exponential);
}

void Test_Estimation::applyDistribution(const TestDistribution *td)
{
	double estimate = _est->estimationType() == Estimation::MEAN ? td->mean : td->quantile;
	unsigned int runs = _est->estimationType() == Estimation::MEAN ? 0 : _RUNS_ - 1;
	
	unsigned int fails = 0;
	unsigned int pnts = 0;
	
	while(runs < _RUNS_)
	{
		while(_error > _prec_)
		{
			_est->advance(td->nextValue());

			if(_est->checkPoint())
			{
				_error = _est->error();
				pnts++;

				//std::cerr << "MEAN: " << _est->estimate() 
				//	<< ", ERROR: " << _error << std::endl;
			}
		}

		double delta = _est->estimate() * _prec_;
		double lower = _est->estimate() - delta;
		double upper = _est->estimate() + delta;
		
		//std::cerr << "0.5 within [" << lower << "," << upper << "]";
		
		if((estimate < lower) || (estimate > upper)) fails++;
		
		//std::cerr << _est->estimate() << " (" << estimate << ")" << std::endl;
		
		_est->reset();
		_error = DBL_MAX;

		runs++;
	}
	
	//std::cerr << "COUNT: " << runs << ", " << "FAILS: " << fails 
	//	<< ", CHKPNTS: " << pnts << std::endl;

	CPPUNIT_ASSERT( (_est->estimationType() != Estimation::MEAN) || (fails < _fmax ));
	CPPUNIT_ASSERT( (_est->estimationType() != Estimation::QUANTILE) || (fails == 0 ));
}
