#ifndef __VARIABLE_H__
#define __VARIABLE_H__

#include <environment/analysis.h>

class Estimation;
class InitialTransient;

//! Messvariable fuer statistische Auswertung.
/*!
 * Ein Messvariable verwaltet die Daten einer Messung und fuehrt eine
 * Auswertung bezueglich Mittelwert oder p-Quantil durch. In stationaren
 * Simulationen wird die Laenge der initialen transiente Phase bestimmt und 
 * die Messwerte dieser Phase verworfen.
 *
 * @see CINSimAnalysis
 * @see Estimation
 * @see InitialTransient
 *
 * @author Matthias Kuehm <kuehmi@cs.tu-berlin.de>
 */
class Variable
{

public:

	//! Zustand der Auswertung.
	enum VarState{TRANS_DETECT, ESTIMATION, FINISHED};
	
	//! Konstruktor.
	/*!
	 * Dier Konstruktor erzeugt eine Messvariable fuer die ueber \a info
	 * spezifierte Messung.
	 * 
	 * @see Analysis::VariableInfo
	 */
	Variable(const Analysis::VariableInfo& info);

	//! Destruktor.
	/*!
	 * Klassenobjekte fuer Auswertung und Erkennung der transienten Phase 
	 * werden freigegeben. 
	 */
	~Variable();
	
	//! Neuen Messwert verarbeiten.
	/*!
	 * Die Verarbeitung eines Messwertes erfolgt in Abhaengigkeit vom Zustand
	 * der Auswertung. Ueber einen Funktionspointer wird die entsprechende
	 * Methode aufgerufen.
	 * 
	 * @see advanceTransient
	 * @see advanceEstimation
	 * @see advanceFinished
	 * 
	 * @param[in] value Neuer Messwert fuer diese Messvariable.
	 * @param[out] result Ergebnisdatenstruktur fuer Messung.
	 * @return true, falls Ergebnis aktualsiert wurde.
	 */
	inline bool advance(double value, Analysis::Result& result)
	{
		return (this->*advanceInternal)(value, result);
	}
	
	//! Aktueller Zustand der Variable.
	/*!
	 * Die von einer Variable verwaltete Auswertung von Simulationsdaten kann
	 * sich entweder in der Erkennung der initialen transienten Phase oder der 
	 * Abschaetzung befinden. Ein weiterer Zustand markiert die Auswertung als
	 * abgeschlossen.
	 * 
	 * @see VarState
	 * @return Aktueller Zustand.
	 */
	inline int state() 
	{ 
		return _state; 
	}
	
	//! Analyse als abgeschlossen markieren.
	/*!
	 * Wenn eine Messung abgeschlossen ist und keine neuen Werte mehr verarbeitet 
	 * werden sollen, so wird ueber diese Methode der Zustand der Messvariable 
	 * aktualiert auf Variable::FINISHED gesetzt und die Methode advanceFinished 
	 * fuer alle weiteren Messwerte verwendet.
	 *
	 * @see advanceFinished
	 */
	inline void markFinished()
	{
		_state = FINISHED;
		advanceInternal = &Variable::advanceFinished;
	}
		
	
protected:
	
	//! Verarbeiten eines Messwertes in transienter Phase.
	bool advanceTransient(double value, Analysis::Result& result);
	
	//! Verarbeiten eines Messwertes bezueglich Auswertung.  
	bool advanceEstimation(double value, Analysis::Result& result);
	
	//! Verarbeiten eines Messwertes nach Abschluss der Auswertung.
	bool advanceFinished(double value, Analysis::Result& result);
	
	//! Funktionspointer fuer Auswertungsmethode.
	bool (Variable:: *advanceInternal)(double, Analysis::Result&);


private:
	
	//! Zustand der Auswertung.
	VarState _state;
	
	//! Auswertung bezuglich Mittelwert oder p-Quantil.
	Estimation* _est;
	
	//! Erkennung der transienten Phase.
	InitialTransient* _itd;
};


#endif
