#include <gnuplot.h>
#include <sequencer.h>
#include <simulation.h>
#include <configuration.h>
#include <analyser.h>

#include <utility/xmlhelper.h>

#include <cmath>
#include <cassert>

void GnuplotOutput::generate()
{	
	Gnuplot* gp = NULL; 
	unsigned int nofana = 0;
	unsigned int nofsim = 0;
	char cbuffer[64];
	
	assert(getDocument() != NULL);

	// Gnuplot vorbereiten
	try 
	{
		gp = new Gnuplot("linespoints");
	}
	catch(GnuplotException e) { throw OutputException(e.what()); }

	gp->cmd("set terminal postscript color");
	gp->cmd(("set output \"" + _file + "\"").c_str());
	gp->cmd("set grid");
	gp->set_ylabel("estimate");		

	// Titel - Dateiname oder Vorgabe aus XML
	stringstream title_cmd;
	title_cmd << "set title \""; 

	const Sequencer::Setup& sim_parameters = Sequencer::Instance()->setup();
	const AnalyserVector& analysers = Simulation::Instance()->Active()->Analysers();

	nofsim = Sequencer::Instance()->Run() + 1;
	nofana = analysers.size();
	
	if(sim_parameters.sname == "-" || sim_parameters.sname.empty())
		title_cmd << XMLString::transcode(getDocument()->getDocumentURI());
	else
		title_cmd << sim_parameters.sname;

	title_cmd << "\"";

	gp->cmd(title_cmd.str().c_str());
			
	// Steady
	if(  Simulation::Instance()->Type() == Simulation::STEADY_STATE )
	{
		// Einzelner Lauf
		if (nofsim == 1)
		{
			assert(Size() == nofana);
			
			throw OutputException("No gnuplot output for single steady state simulation runs!");
		}
		
		// Sequenz
		else if (nofsim > 1)
		{
			assert(Size() == nofana * nofsim);

			std::vector<double> x;	// x axis values
			std::vector<double> y;	// y axis values
			double ymax = 0.0;		// Maximum fuer y
			
			gp->set_xlabel("simulation run");
			gp->cmd("set xrange [1:]");
			// Wertebereich setzen
			gp->cmd("set yrange [0:]");
			
			// alle Variablen
			for (unsigned int ana=0; ana < nofana; ana++)
			{							
				for (unsigned int run=0; run < nofsim; run++) 
				{
					x.push_back(run+1);
					y.push_back(getResult(run*nofana+ana).Estimate());

					ymax = ymax < y.back() ? y.back() : ymax;
				}
							
				// Wertebereich setzen
				sprintf(cbuffer,"set yrange [0:%f]", ceil(ymax));
				gp->cmd(cbuffer);

				try 
				{
					// Plot
					gp->plot_xy(x, y, (analysers[ana]->Description() + " (" 
							+ analysers[ana]->Measurement() + ")").c_str());
				}
				catch(GnuplotException e) { throw OutputException(e.what()); }

				gp->reset_plot();
					
				// Werte zuruecksetzen
				x.clear();
				y.clear();
				ymax = 0.0;
				
				*out << "++ Created plot for variable " << ana+1 << std::endl;

			} // mvars
		}
		else throw OutputException("Panic! Unexpected Error!");
	}
	
	// Transient
	else if(sim_parameters.stype == "Terminating") 
	{
		unsigned int offset = sim_parameters.tend-sim_parameters.tstart+1;
		
		assert(Size() == nofana * offset * nofsim);

		std::vector< std::vector<double> > rx; // x axis values
		std::vector< std::vector<double> > ry; // y axis values
		double ymax;

		gp->set_xlabel("clock cycle");
		sprintf(cbuffer, "set xrange [%d:]",sim_parameters.tstart);
		gp->cmd(cbuffer);
		
		// alle Variablen
		for (unsigned int ana=0; ana < nofana; ana++)
		{			
			gp->cmd("set multiplot");
											
			// Werte fuer alle Laeufe sammeln
			rx.resize(nofsim);
			ry.resize(nofsim);
			ymax = 0.0;

			for (unsigned int run=0; run < nofsim; run++) 
			{
				for (unsigned int m=0;m<offset;m++) 
				{
					rx[run].push_back(sim_parameters.tstart+m);
					ry[run].push_back(getResult(run*nofana*offset+ana*offset+m).Estimate());

					ymax = ymax < ry[run].back() ? ry[run].back() : ymax;
				}
			}
			
			// Wertebereich setzen
			sprintf(cbuffer,"set yrange [0:%f]", ceil(ymax));
			gp->cmd(cbuffer);

			// Plot
			for (unsigned int run=0; run < nofsim; run++)
			{
				sprintf(cbuffer, "%s, Run %d", (analysers[ana]->Description() + " (" 
						+ analysers[ana]->Measurement() + ")").c_str(), 
						run+1);
				try
				{
					gp->plot_xy(rx[run], ry[run], cbuffer);
				}
				catch(GnuplotException e) { throw OutputException(e.what()); }
			}
			
			gp->cmd("unset multiplot");
			gp->reset_plot();
				
			// Werte zuruecksetzen
			rx.clear();
			ry.clear();

			*out << "++ Created plot for variable " << ana+1 << std::endl;

		} // measures
	}
	else
		throw OutputException("Panic! Unknown simulation type!");

	*out << "++ Plot(s) successfully written to " << _file << std::endl;
}


