#include <router.h>
#include <targetbuffer.h>
#include <utility/xmlhelper.h>
#include <macros.h>
#include <packetforwarding.h>
#include <configuration.h>
#include <switching.h>
#include <wirematrixiterator.h>
#include <packet.h>
#include <registercontainer.h>
#include <bufferqueue.h>

#include <cassert>

REGISTER_ELEMENT(Router, "router")


bool Router::init(const DOMElement* scope) 
{	
	if(_ac) return true;
	
	if(scope) setScope(scope);
	
	assert(Scope());
	
	_ac = (ArbiterController*)FactoryElement::Factory::self().
			create(conv(Scope()->getAttribute(conv("controllerType"))));

	assert(_ac);
	
	_ac->init(Scope());
	_ac->setMatrix(&_wm);

	// Scheduler fuer jeden Ausgang
	for (unsigned int i=0; i< _wm.getOutWires().size(); i++)
		_ac->addArbiter();

	return true;
}


const Bitmask& Router::generateBitmask() 
{
	// we did this call allready
	if (_seen) return _bitmask;

	// Gegen boese Zyklen
	_seen = true;
	
	for (unsigned int r=0; r<Out().size(); r++)
		_bitmask |= Out()[r]->generateBitmask();

	_seen = false;
	return _bitmask;
}


void Router::reset()
{
	unsigned int i=0;
	WireMatrixIterator* it=NULL;
	
	// Schaltungen aufheben
	for( ; i<Out().size(); i++)
	{
		it=_wm.iterateInWires(Out()[i]);

		// Bits zuruecksetzen
		while(it->hasNext())
		{
			it->next();
			it->value().reset();			
		}

		// ausgehende Leitung freigeben
		Out()[i]->unconnectRegister();
	}

	// Informationen zu anliegenden Paketen ungueltig
	for(i=0; i<_wm.managedPackets().size(); i++)
	{
		_wm.managedPackets()[i].valid=false;
	}
}

// RouterMatrix aus Sicht eines Ausgangs aktivieren
// - umfasst Scheduing, Switching
int Router::putPacketOnLine(Wire* client, bool& head)
{
	// Annahme: Kein Router-Router
	assert(client->To()->rtti()<=Component::TARGET);
	// Annahme: Noch keine Schaltung auf diesen Ausgang
	assert(!client->RegisterOnLine());
	
	WireMatrixIterator* wmi = _wm.iterateInWires(client);
	bool scheduled=false;
	ModifierList connect;
	
	// noch keine Entscheidung fuer Ausgang
	if(!wmi->value().test(O_SCHEDULED))
	{
		_DEBUG(toString() << " : Matrix pre ac : " << _wm);
		scheduled = _ac->activate(client);
		_DEBUG(toString() << " : Matrix post ac : " << _wm);	
		
		if(!scheduled)
			return TriggerNode::WAIT;

		// Entscheidung in Matrix eintragen
		connect.push_back(_ac);
	}
	
	// Entscheidung fuer Ausgang erfordert neue Schaltung
	if(!wmi->value().test(O_FINISHED))
	{
		// Ausgang schalten
		connect.push_back(Coroutine()->switching());
	}
		
	// Schaltmatrix aktualisieren
	wmi->setAll(&connect);

	// Flit liegt an
	
	if(client->RegisterOnLine())
	{
	    assert(client->getFlit(client->To()));
	
	    head = client->getFlit(client->To())->head();
	    
	    return TriggerNode::SUCCESS;
	}
	
	return TriggerNode::NOPACKET;
}

// RouterMatrix aus Sicht eines Eingangs initialisieren
// - umfasst PreSwitching, Routing
int Router::lineActivated(Wire* inwire, ModifierList* prepare, 
		SignedTNodeList* list, TNodeList::iterator pos)
{
	// Annahme: Leitung ist eingehende Leitung dieses Routers
	assert(inwire->getInIndex() < In().size() && In()[inwire->getInIndex()] == inwire);
	// Annhame: WM besitzt Feld fuer zu vermittelndes Paket
	assert(_wm.managedPackets().size() > inwire->getInIndex());
	
	// Zuerst muessen alte Zustaende entfernt werden
	prepare->push_front(&_preswitch);
	
	// Wird ein Paket (Head) zum ersten mal angelegt, so werden Informationen
	// fuer die Schaltmatrix generiert
	WireMatrix::PacketInfo* managed = &_wm.managedPackets().begin()[inwire->getInIndex()];
	const Packet* online = inwire->getPacket();	
	if(online)
	{
		if((!managed->valid || managed->number != online->number))
		{
			managed->packet = online;
			managed->number = online->number;
			
			//! Annahme: Eina alte Bitmaske existiert
			assert(managed->remain);
			
			delete managed->remain;

			managed->remain = new Bitmask(online->getBitmask());
			managed->valid=true;
			
			//! Annahme: Bitmaske mit verbleibenden Zielen ist nicht leer.
			assert(!managed->remain->empty());
		}

		// Annahme: Paket und ein Bitmaske sind registriert
		assert(managed->packet && managed->remain);
	}
	else
		managed->valid=false;
	
	// PreSwitching + Routing (optional)
	WireMatrixIterator* wmi = _wm.iterateOutWires(inwire);
	wmi->setAll(prepare);
	
	// moegliche Empfaenger werden verlangt
	if(list && online)
	{
		wmi->reset();
		while(wmi->hasNext())
		{
			wmi->next();
			WireMatrixEntry entry = wmi->value();
			if(entry.test(IO_VALID))
				list->insert(pos,(TriggerNode*)entry.getOutWire()->To());
		}
	}
	return TriggerNode::SUCCESS;
}

// Schaltung aus Sicht eines Ausganges abschliessen
// -> fertige bearbeitete Eingaenge werden freigegeben
int Router::lineFinished(Wire* outwire, ModifierList* clear,
			SignedTNodeList* list,TNodeList::iterator pos)
{
	TNodeList::iterator front = pos;

	// Annahme: Leitung ist ausgehende Leitung dieses Routers
	assert(outwire->getOutIndex() < Out().size() && Out()[outwire->getOutIndex()]==outwire);

	WireMatrixEntry entry;
	Wire* inwire = NULL;
	WireMatrixIterator* wmi = _wm.iterateInWires(outwire);
	
	// Annahme: Ausgang fertig bearbeitet
	assert(wmi->value().test(O_FINISHED));
	
	// Schaltung soll aufgehoben werden
	if(clear)
	{
		wmi->setAll(clear);
		wmi->reset();
	}
	
	while(wmi->hasNext())
	{
		wmi->next();
		entry = wmi->value();
			
		// nur relevante Eingaenge betrachten
		if(!entry.test(IO_VALID))
			continue;
		
		// Eingang zu Matrixeintrag
		inwire = entry.getInWire();

		// Geschalteter Eingang
		if(entry.test(IO_WON))
		{
			// Annahme: Schaltung fuer Eingang und Ausgang markiert
			assert(entry.test(IO_SWITCHED));
					
			// fertig bearbeitete Knoten einsammeln
			inwire->From()->lineFinished(inwire,clear,list,front);
						
			// weiter mit naechstem Eingang
			continue;
		}
		
		// Annahme: nicht abgeholtes Flit noch auf der Line
		assert(inwire->getFlit(entry.getOutWire()->To()));
		
		_PACKETMOVE("Rejected flit:",inwire->getFlit(entry.getOutWire()->To())->toString(),
				"From: " << inwire->From()->toString());

		// Annahme: Register auf der Line 
		assert(inwire->RegisterOnLine());
		
		// Eingang konnte kein Flit an diesen Ausgang vermitteln.
		inwire->RegisterOnLine()->rejectFlit(entry.getOutWire()->To());
		
		// fertig bearbeitete Knoten einsammeln
		inwire->From()->lineFinished(inwire,clear,list,pos);
		if(front!=NLITNULL) front--;
	}			
	
	return TriggerNode::SUCCESS;
}

// Hinzufuegen einer eingehenden Leitung
unsigned int Router::addInWire(Wire* wire)
{
	unsigned int index = Node::addInWire(wire);
	
	// WireMatrix aktualisiern
	_wm.newInWire();
		
	return index;
}

// Hinzufuegen einer ausgehenden Leitung
unsigned int Router::addOutWire(Wire* wire)
{	
	unsigned int index = Node::addOutWire(wire);

	// WireMatrix aktualisiern
	_wm.newOutWire();

	return index;
}

// Zuruecksetzen alter Schaltzustaende aus letztem Takt
void PreSwitching::modify(WireMatrixIterator* it)
{
	WireMatrixEntry& entry = it->value();
	WireMatrixEntry mask;
	
	// Eingang bleibt auf Ausgang geschaltet
	if(entry.test(I_BUSY))
	{
		// Annahme: Schaltung auf Ausgang
		assert(entry.test(IO_SWITCHED));
		assert(entry.test(IO_WON));
		assert(entry.test(IO_VALID));
		assert(!entry.test(IO_LOST));
		assert(entry.test(O_BUSY));
		
		// Schaltung bleibt		
		mask.set(IO_SWITCHED);
		mask.set(I_BUSY);
		mask.set(I_FINISHED);
				
		// Schedulingergebnis bleibt
		mask.set(IO_WON);
	}
		
	// Ausgang bleibt geschaltet  
	if(entry.test(O_BUSY))
	{
		// Annahme: Ausgang als geschaltet markiert
		assert(entry.test(O_SCHEDULED));
		assert(entry.test(O_FINISHED));
		
		// Schaltung bleibt
		mask.set(O_BUSY);
		mask.set(O_SCHEDULED);
		mask.set(O_FINISHED);

		// Schedulingergebnis bleibt
		mask.set(IO_LOST);
	}
	
	// Annahme: Wenn Flit fuer Ausgang verfuegbar, dann bereits geroutet
	assert(!entry.getInWire()->getFlit(entry.getOutWire()->To()) || entry.test(IO_VALID));
	
	// Routing bleibt bestehen solange Flit verfuegbar
	mask.set(IO_VALID, entry.getInWire()->getFlit(entry.getOutWire()->To()));
	
	// Matrixentrag aktualisieren
	entry &= mask;

	// Annahme: Wenn Paket fuer Matrix registriert, dann auch auf Line
	assert(!entry.currentPacket() || (entry.getInWire()->getPacket()==entry.currentPacket())); 

	// Bit fuer eventuell anliegendes Paket setzen
	entry.set(I_PACKET, entry.currentPacket());

	// Annahme: Wenn Routingentscheidung vorliegt, dann auch Paket
	assert(!entry.test(IO_VALID) || entry.test(I_PACKET));
}

