#ifndef __WFROUTING__
#define __WFROUTING__

#include <routing.h>
#include <utility/bitmask.h>

#include <vector>

//! West-First-Routing fuer Gitternetze (DyNoC).
/*!
 * Diese Klasse implementiert ein West-First-Routing fuer Gitternetze. Pakete 
 * werden zuerst in Richtung Westen (x wird kleiner) bewegt bis mindestens
 * die korrekte X-Koordinate erreicht ist, dann adaptiv nach Norden/Sueden
 * (y wird kleiner/groesser) oder Osten (x wird kleiner). Ein Richtungswechsel
 * nacht Westen ist nicht erlaubt.
 * 
 * Die Implementierung betrachtet zwei Phasen: In der initialen Phase 1 sind
 * Bewegungen nach Westen moeglich und fuer Ziele im Westen auch zwingend. In
 * Phase 2 wird angenommen, dass nur noch Ziele in Norden, Sueden oder Osten
 * erreicht werden sollen.
 *
 * Hinweis: Voraussetzung ist, dass Netzknoten ueber entsprechende 
 * Positionsinformationen verfuegen. Sind diese nicht vorhanden, so findet 
 * kein Routing statt. Weiterhin muessen zu simulierende Gitternetze vollstaendig
 * sein und eine rechteckige Grundform aufweisen.
 * 
 * @author Matthias Kuehm <kuehmi@cs.tu-berlin.de>
 */
class WFRouting : public Routing
{
public:

	//! Implementierung der abstrakten Basisklasse Routing.
	/*!
	 * Bei der Initialisierung des West-First-Routings wird das zu simulierende
	 * Gitternetz untersucht und ausgehend von den Bitmasken der Puffer 
	 * im Netz, die die theoretisch von einem Puffer erreichbarene Ziele
	 * beschreiben, eingeschraenkte Bitmasken generiert, die die zulaessigen
	 * Ziele zu jedem Puffer aus Sicht eines schaltenden Routers beschreiben.
	 * 
	 * @see Routing
	 */
	bool init();
		
protected:
		
	//! Implementierung der abstrakten Basisklasse Routing.
	/*!
	 * Unterscheidet sich die aktuelle Position von der Position
	 * eines Ziels in der X-Koordinate, so ist nur ein Ausgang
	 * zulaessig, der diese Differenz verkleinert. Ist die
	 * korrekte X-Koordinate erreicht, so wird analog fuer
	 * die Y-Koordinate verfahren.
	 *
	 * @see Routing
	 */
	Bitmask* validOutput (const Packet* ptr, const Bitmask* remaining, const Wire* output);

private:
		
	//! Bitmasken zulaessiger Ziele fuer alle Puffer im Netz (Phase 1)
	std::vector<Bitmask> _bitmasks_p1;

	//! Bitmasken zulaessiger Ziele fuer alle Puffer im Netz (Phase 2)
	std::vector<Bitmask> _bitmasks_p2;
};

#endif
