#include <pareto.h>
#include <utility/xmlattribute.h>
#include <macros.h>
#include <environment/prng.h>

#include <math.h>
#include <cfloat>

REGISTER_ELEMENT(ParetoDistribution, "Pareto")

void ParetoDistribution::reset()
{
	burstTick = 0;
	gapTick = 0;
	burstInterval = 0;
	gapInterval = 0;
	if( UniformDouble_1() > 0.5 ) state=BURST; else state=GAP; 
}

bool ParetoDistribution::init(const DOMElement* scope)
{           
	if(scope)
		setScope(scope);

	if(!Scope())
		_ERROR("Cannot initialize traffic distribution without its XML element!");

	load = getAttribute("load");
	onShape = getAttribute("onShape");
	offShape = getAttribute("offShape");
	constPacketSize= getAttribute("packetSize", getParent());
	
	if(constPacketSize != 1) 
		_ERROR("[ParetoDistribution] Packet sizes larger than 1 are not supported yet.");                                
	if(load < 0.0 || load > 1.0)
		_ERROR("[ParetoDistribution] Load (" << load << ") out of range!");

	if(onShape<=1 || onShape>=2) 
		_WARNING("[ParetoDistribution] Specified attribute onShape not between 1 and 2!");

	if(offShape<=1 || offShape>=2)
		_WARNING("[ParetoDistribution] Specified attribute offShape not between 1 and 2!");

	// minimale Intervallgroessen
	minBurst = 1;
	minGap = calculateCoef();

	// initialer Zustand
	reset();
	
	return true;
}

bool ParetoDistribution::ifNowGenerate(unsigned int)
{
	advanceState();
	
	return state;
}

inline double ParetoDistribution::Xpareto(double shape) const
{
	return (pow( UniformDouble(), -1.0/shape));
}

void ParetoDistribution::advanceState()
{
	if (state == BURST)
	{
		if (burstTick >= burstInterval) 
		{
			state = GAP;
			gapTick = 1;
			burstTick = 0;
			burstInterval = 0;
			
			gapInterval = (unsigned int) rint((Xpareto( offShape ) * minGap));
			
			//std::cerr << "GAP: " << gapInterval << " (" << minGap << ")" << std::endl;

			if(gapInterval > 0) return;
			// if gapInterval==0, a new burst will be set up
		} 
		else burstTick++;
	} 
	
	if (state == GAP) 
	{
		if (gapTick >= gapInterval) 
		{
			state = BURST;
			gapTick = 0;
			gapInterval = 0;
			burstTick=1;
			
			burstInterval = (unsigned int) rint((Xpareto( onShape ) * minBurst ));	
		
			if (burstInterval < minBurst)  
			{
				burstInterval = minBurst;
			}
			
			//std::cerr << "BURST: " << burstInterval << std::endl;
		} 
		else gapTick++;
	}
}

double ParetoDistribution::calculateCoef() const
{
	double on_coef, off_coef, on_scoef, off_scoef, coef;

	on_coef  =   ( onShape - 1.0 )/onShape ;
	off_coef =   ( offShape - 1.0 )/offShape;
	on_scoef =  pow(DBL_MIN,  on_coef);
	off_scoef =  pow(DBL_MIN,  off_coef);
	coef = off_coef / on_coef * (1.0 / load - 1.0);
	return (1.0 - on_scoef)/(1.0 - off_scoef) * coef;
}
