#ifndef __RANDOM_BURST__
#define __RANDOM_BURST__

#include <traffic.h>

//! Verkehr mit zufaelligen Aktivitaets- und Ruhephasen.
/*!
 * Diese Paketverteilung wird durch den Wechsel von Aktivitaets- und
 * Ruhephasen mit entsprechenden Paketraten charakterisiert. Die 
 * Laenge der einzelnen Phasen wird zufaellig bestimmt.
 * 
 * @author Li Zhou
 */
class RandomBurstsDistribution : public TrafficDistribution
{
public:

	//! Konstruktor.
	/*!
	 * Der Konstruktor initialisiert alle Klassenattribute der Basisklassen
	 * und alle nicht vom Benutzer zu spezifizierenden Attribute.
	 *
	 * @see TrafficDistribution
	 * @see XMLElement
	 * 
	 * @param psize Groesse der zu generierenden Pakete.
	 * @param scope Gueltigkeitsbereich innerhalb des XML-Dokumentes.
	 */	
    RandomBurstsDistribution(unsigned int psize=1, const DOMElement* scope=NULL)
		: TrafficDistribution(psize, scope) {}
	
	//! Implementierung der Basisklasse XMLElement.
	/*!
	 * Die Verteilung wird anhand der relavanten Attribute aus dem
	 * Gueltigkeitsbereich innerhalb des XML-Dokumentes initialisiert.
	 *
	 * @see XMLElement.
	 * 
	 * @param scope Der Gueltigkeitsbereich.
	 * @return true.
	 */
	bool init(const DOMElement* scope=NULL);

	//! Implementierung der Basisklasse TrafficDistribution.
	/*!
	 * Pakete koennen innerhalb von Aktivitaets- und Ruhephasen mit
	 * unterschiedlichen Raten generiert werden.
	 * 
	 * @param tick Der aktuelle Takt.
	 * @return true, falls laut Verteilung Paket generiert werden kann.
	 */
    virtual bool ifNowGenerate(unsigned int tick);

	//! Implementierung der Basisklasse TrafficDistribution.
	/*!
	 * Alle Klassenattribute, die den internen Zustand der Verteilung
	 * beschreiben, werden zurueckgesetzt.
	 */
	void reset();
	
	//! Destruktor.
	~RandomBurstsDistribution() {};
	
protected:
	
	//! Zustandsuebergang.
	/*!
	 * Diese Methode veraendert den internen Zustand dieser Verteilung
	 * derart, dass eine Folge von Aktivitaets- und Ruhephasen entsteht.
	 */
	void advanceState();

private:
	
	// -- Benutzerangaben --
	
	double BurstLoad;					//User defines inputrate in burstinterval
	double GapLoad;
	unsigned int BurstIntervalMin;		//User defines minimum burstinterval
	unsigned int BurstIntervalMax;
	unsigned int GapIntervalMin;		//User defines minimum gapinterval
	unsigned int GapIntervalMax;
	
	// -- interner Zustand --
		
	unsigned int burstTick;			//elapsed bursttime
	unsigned int gapTick;
	unsigned int burstInterval;		//Random generated
	unsigned int gapInterval;
	enum { GAP, BURST }state;
};

#endif
