#include <xmlsimulation.h>
#include <environment.h>
#include <node.h>
#include <wire.h>
#include <analyser.h>
#include <configuration.h>
#include <macros.h>

#include <utility/xmlhelper.h>
#include <environment/analysis.h>

#include <cassert>


using namespace std;



void XmlSimulation::updateConnectables(XMLElement *xmlelem, Connectables *map) const
{
	const DOMElement* sibling = xmlelem->Scope();
	bool root = true;

	while(sibling) 
	{
		if(sibling->getNodeType() == DOMNode::ELEMENT_NODE)
		{
			assert(_schemaInformation->getValue(conv(sibling->getTagName())));
		
			if(root || isDerivedFrom(_schemaInformation, sibling, "IOType"))
			{
				(*map)[parseAttribute(sibling, "id")] = xmlelem;
			}

			if(root)
			{
				sibling = (const DOMElement*)sibling->getFirstChild();
				root = false;
				continue;
			}
		}
		sibling = (const DOMElement*)sibling->getNextSibling();
	}
}


void XmlSimulation::updateSettings(DOMElement* elem, Settings *current, const Settings *previous)
{
	DOMTreeWalker* settings = createElementWalker(elem, NULL, false);
	DOMElement *setting = (DOMElement*)settings->firstChild();

	while(setting)
	{
		if(equals(setting->getNodeName(), "simsettings"))
		{
			(*current)["simulation"] = setting;
		}
		
		if(equals(setting->getNodeName(), "buffersettings"))
		{
			(*current)["buffer"] = setting;
		}

		else if(equals(setting->getNodeName(), "routersettings"))
		{
			(*current)["router"] = setting;
		}

		else if(equals(setting->getNodeName(), "sourcebuffersettings"))
		{
			(*current)["sourcebuffer"] = setting;

			// 
			// FIXME: ggf. Node-Iterator aus trunk verwenden?
			//
			(*current)["geometric"] = getFirstChildByName(setting,"geometricsettings");
			(*current)["periodic"] = getFirstChildByName(setting,"periodicsettings");
			(*current)["pareto"] = getFirstChildByName(setting,"paretosettings");
			(*current)["randomburst"] = getFirstChildByName(setting,"randomburstsettings");

			assert(current->find("geometric")->second != NULL);
			assert(current->find("periodic")->second != NULL);
			assert(current->find("pareto")->second != NULL);
			assert(current->find("randomburst")->second != NULL);
		}

		setting = (DOMElement*)settings->nextSibling();
	}
	
	assert(current->find("simulation") != current->end());
	assert(current->find("buffer") != current->end());
	assert(current->find("sourcebuffer") != current->end());
	assert(current->find("router") != current->end());

	// Ziele mit defaults fuer Puffer
	(*current)["targetbuffer"] = (*current)["buffer"];
	// Quellen zusaetzlich mit defaults fuer Puffer
	mergeElements((*current)["buffer"], (*current)["sourcebuffer"]);

	if(previous)
	{
		Settings::iterator cit;
				
		for(cit=current->begin(); cit!=current->end(); cit++)
		{
			//std::cerr << "..." << cit->first << std::endl;
			
			assert(previous->find(cit->first) != previous->end());
			assert(previous->find(cit->first)->second != NULL);

			mergeElements(previous->find(cit->first)->second, (*current)[cit->first]);
		}
	}
}


void XmlSimulation::init()
{	
	vector<DOMElement*> reconfigs;
	vector<DOMElement*> measures;

	Settings defaults;
	Connectables configs;
	
	DOMElement* net = getFirstChildByName(_document->getDocumentElement(), "net");
    
	if(!net)
		_ERROR("Required top level element 'net' not found!");
	
	DOMTreeWalker *walker = createElementWalker(net, NULL, false);
	DOMElement *sibling = (DOMElement*)walker->firstChild();
	while(sibling) 
	{
		if(isNode(sibling))
		{
			if(equals(sibling->getNodeName(), "netglobals")) 
			{
				updateSettings(sibling, &defaults);
			}
		
			else if(isDerivedFrom(_schemaInformation, sibling, "ConfigurationNodeType"))
			{
				Configuration* config = new Configuration(sibling);
				
				configs[config->getAttribute("id")] = config;
			}
		}
			
		else if(isConnector(sibling))
		{
			if(equals(sibling->getNodeName(), "netreconfig")) 
			{
				reconfigs.push_back(sibling);
			}
		}

		else if(isText(sibling))
		{
			if(equals(sibling->getNodeName(), "measurevar")) 
			{
				measures.push_back(sibling);
			}
		}
		sibling = (DOMElement*)walker->nextSibling();
	}
	
	assert(configs.size());

	//
	// Stochastische Auswertung intialisieren
	//
	
	const DOMElement* simsettings = defaults.getValue("simulation");
	Analysis::VariableInfo* list = NULL;
	unsigned int count = measures.size();
	unsigned int length = 1;
	
	if((string)parseAttribute(simsettings, "simulationType") == "Terminating")
	{
		length = parseAttribute(simsettings, "analysisLength");
		
		assert(length>0);
		
		count *= length;
		
		_type = TERMINATING;
		_analysisStart = parseAttribute(simsettings, "analysisStart");
		_analysisEnd = _analysisStart + length - 1;
	}	
	else
	{
		_type = STEADY_STATE;
	}

	list = new Analysis::VariableInfo[count];
	Analysis::VariableInfo tmp;
	unsigned int offset = 0;
	for(unsigned int i=0; i<count; i++)
	{
		if(i%length == 0)	
		{
			string type = parseAttribute(measures[offset], "measurementType");
						
			if(type == "Mean")
				tmp.type = Analysis::VariableInfo::Mean;
			else if(type == "Quantile[p]")
				tmp.type = Analysis::VariableInfo::Quantile;
			else
				_ERROR("Unknown type for measurement ( " << type << " )!");
					
			if(_type == STEADY_STATE)
			{
				tmp.independent = false;
				tmp.initialtransient = true;
			}
			else
			{
				tmp.independent = true;
				tmp.initialtransient = false;
			}

			if(measures[offset]->hasAttribute(conv("confidence")))
				tmp.confidence = parseAttribute(measures[offset], "confidence");
			else
				tmp.confidence = Env()->getValue(_CONFIDENCE_);

			if(measures[offset]->hasAttribute(conv("precision")))
				tmp.precision = parseAttribute(measures[offset], "precision");
			else
				tmp.precision = Env()->getValue(_PRECISION_);

			tmp.quantile = parseAttribute(measures[offset], "p");
			
			offset++;
		}
		
		//std::cerr << "DEBUG: i=" << i << ", offset=" << offset 
		//		<< ", type=" << tmp.type << ", p=" << tmp.quantile 
		//		<< ", prec=" << tmp.precision 
		//		<< ", conf=" << tmp.confidence << std::endl;;

		list[i] = tmp;
	}
	
	assert(Analysis::Instance());
	Analysis::Declare(list, count);

	//
	// Konfigurationen initialisieren
	//

	for(Connectables::iterator it = configs.begin(); it!=configs.end(); it++)
	{
		Configuration *config = (Configuration*)it->second;
			
		for(unsigned int i=0; i<measures.size(); i++)
		{
			config->_analysers.push_back(new Analyser(measures[i]));
		}
		
		config->_settings = defaults.getValue("simulation");
		
		assert(config->_settings);
		assert(config->Scope());
		
		initConfig(config->Scope(), &defaults, config);
	
//#ifndef NDEBUG
		vector<Component*> &wc = config->_components[Component::WIRE]; 
		for(unsigned int j=0; j<wc.size(); j++)
		{
			//assert(((Wire*)wc[j])->From());
			//assert(((Wire*)wc[j])->To());
			if(!((Wire*)wc[j])->From() || !((Wire*)wc[j])->To())
				_ERROR("Corrupted component seen: " << wc[j]->toString())
		}
//#endif
		_wires.clear();

		// Konfiguration in Simulation aufnehmen
		config->_index = _configurations.size();
		_configurations.push_back(config);
		
		if((string)config->getAttribute("initialConfig") == "true")
		{
			if(_initialConfig)
				_WARNING("More than one initial configuration found! "
						<< "The first one (" 
						<< (string)_initialConfig->getAttribute("index") 
						<< ") will be used!")
				
			else
				_initialConfig = config;
		}
		
		// Konfiguration initialisieren
		_activeConfig = config;
		config->init();
	}

	if(_configurations.size() == 0)
		_ERROR("No network configuration found!")

	else if(_configurations.size() == 1)
		_initialConfig = _configurations[0];

	else if(_configurations.size() > 1 && !_initialConfig)
		_ERROR("No initial network configuration selected!")
		
	//
	// Rekonfigurationsfolge bestimmen
	//
	
	unsigned int drainLenDyn = ((int)parseAttribute(defaults.getValue("buffer"), "minSpace")) - 1;

	for(unsigned int i=0; i<reconfigs.size(); i++)
	{
		Configuration* from = 
			(Configuration*)configs.getValue(parseAttribute(reconfigs[i], "fromNode"));
		Configuration* to =
			(Configuration*)configs.getValue(parseAttribute(reconfigs[i], "toNode"));

		assert(from && to);
		
		from->_operationPhase = parseAttribute(reconfigs[i], "operationPhase");
		from->_drainPhase = parseAttribute(reconfigs[i], "packetDrainPhase");
		from->_reconfigPhase = parseAttribute(reconfigs[i], "reconfigurationPhase");

		from->_dynamic = 
			((string) parseAttribute(reconfigs[i], "dynamic"))=="false" ? false : true;

		//
		// Erfassen, ob Virtual Cut-Through bzw. Wormhole Switching in Verbindung
		// mit dynamischer Rekonfiguration gewaehlt wurden -> Abbruch
		//
		
		if(from->_dynamic 
			&& (((string)parseAttribute(from->Settings(),"switchingType")) == "VirtualCutThrough"
			|| ((string)parseAttribute(to->Settings(),"switchingType")) == "VirtualCutThrough")) 
			_ERROR("Virtual cut-through switching is not supported with dynamic reconfiguration.");

		if(from->_dynamic
			&& (((string)parseAttribute(from->Settings(),"switchingType")) == "Wormhole"
			|| ((string)parseAttribute(to->Settings(),"switchingType")) == "Wormhole")) 
			_ERROR("Wormhole switching is not supported with dynamic reconfiguration.");
		
		
		// Wenn dynamic, dann ist _drainPhase eine collectPhase mit Lnge PL-1
		if(from->_dynamic) from->_drainPhase = drainLenDyn;
		
		from->_nextConfig = to;
		
		//
		// Nun werden die Tabellen erstellt, wo in jeder Konfiguration die Buffer- und
		// Target-Nachfolger eingetragen sind.
		// 
		
		std::cout << "# Set up tables in xmlsimulation.cpp! #\n";

		ComponentVector fromBuffers = from->Components(Component::BUFFER);
		ComponentVector toBuffers = to->Components(Component::BUFFER);
		from->_nextBuffer.clear();
		for(unsigned int ii = 0;ii < fromBuffers.size();ii++)
		{
			from->_nextBuffer.push_back(-1); // zuerst einmal kennt kein Buffer seinen Nachfolger
		}

		ComponentVector fromTargets = from->Components(Component::TARGET);
		ComponentVector toTargets = to->Components(Component::TARGET);
		from->_nextTarget.clear();
		for(unsigned int ii = 0;ii < fromTargets.size();ii++)
		{
			from->_nextTarget.push_back(-1); // und auch kein Target den seinen
		}

		// Buffer-Tabelle fllen, die die entsprechenden Bufferindizes in _nextConfig wissen soll.

		// fr alle toBuffers it2
		for(ComponentVector::iterator it2 = toBuffers.begin();it2!=toBuffers.end();it2++)
		{
			// finde jeden Clone
			if((string)(*it2)->getAttribute("cloned") == "true")
			{ 
				// und suche dessen fromBuffer it1 
				for(ComponentVector::iterator it1 = fromBuffers.begin();it1!=fromBuffers.end();it1++)
				{  
					// wenn gefunden, dann Tabellenzeile in _nextBuffer einfllen
					if((*it1)->InternalName() == (*it2)->InternalName()) 
					{
						from->_nextBuffer[(*it1)->Index()] = (*it2)->Index();
						break;
					}
				}
			}
		}

		// Target-Tabelle fllen dito
		
		// fr alle toTargets it2
		for(ComponentVector::iterator it2 = toTargets.begin();it2!=toTargets.end();it2++)
		{
			// finde jeden Clone
			if((string)(*it2)->getAttribute("cloned") == "true")
			{ 
				// und suche dessen fromTarget it1
				for(ComponentVector::iterator it1 = fromTargets.begin();it1!=fromTargets.end();it1++)
				{  
					// wenn gefunden, dann Tabellenzeile in _nextTarget einfllen
					if((*it1)->InternalName() == (*it2)->InternalName()) 
					{
						from->_nextTarget[(*it1)->Index()] = (*it2)->Index();
						break;
					}
				}
			}
		}

		// Tabellen ausgeben
		/*
		std::cout << "Bold Bnew\n";
		for(unsigned int i=0;i<fromBuffers.size();i++)
			std::cout << "  " << i << "   " << from->_nextBuffer[i] << "\n";

		std::cout << "\nTold Tnew\n";
		for(unsigned int i=0;i<fromTargets.size();i++)
			std::cout << "  " << i << "   " << from->_nextTarget[i] << "\n";
		*/
	}
}


void XmlSimulation::initConfig(const DOMElement* elem, const Settings *previous, Configuration *config)
{
	Connectables nodes;
	Settings defaults;
	const Settings* defptr = NULL;
	StringMap< vector<XMLElement*> > measurements;
	vector<DOMElement*> metacomponents;
	
	DOMElement *net = getFirstChildByName(elem, "net", endsWith);
	
	assert(net);

	DOMTreeWalker *walker = createElementWalker(net, NULL, false);
	DOMElement *sibling = (DOMElement*)walker->firstChild();

	while(sibling)
	{	
		// Netzknoten
		if(isNode(sibling))
		{
			// Einstellungen
			if(isDerivedFrom(_schemaInformation, sibling, "DefaultsType"))
			{
				// Annahme: Einstellungen vor Komponenten
				assert(!defptr);
				
				//std::cerr << "-> Settings: " << StrX(sibling->getTagName()) << std::endl;
				
				updateSettings(sibling, &defaults, previous);
				
				// Einstellungen in Top-Level Netz einer Konfiguration 
				// ueberschreiben globale Einstellungen
				if(isDerivedFrom(_schemaInformation, elem, "ConfigurationNodeType"))
				{
					config->_settings = defaults.getValue("simulation");
				}
				
				defptr = &defaults;
			}
			
			// Metakomponenten
			else if(isDerivedFrom(_schemaInformation, sibling, "MetaComponentType"))
			{
				if(elem->hasAttribute(conv("coords")))
					addAttribute(sibling, elem->getAttributeNode(conv("coords")));
					
				metacomponents.push_back(sibling);
			}
			
			// Komponenten
			else if(isDerivedFrom(_schemaInformation, sibling, "NodeComponentType"))
			{
				if(!defptr) defptr = previous;
				
				Node *node = (Node*)FactoryElement::Factory::self().
					create(conv(sibling->getTagName()));

				// Annahme: Instanz erstellt
				assert(node);
				// Annahme: defaults verfuegbar
				assert(defptr->getValue(conv(sibling->getTagName())));
				
				// defaults fuer fehlende Attribute
				mergeElements(defptr->getValue(conv(sibling->getTagName())), sibling);

				// defaults fuer fehlende Attribute von direkten Subelementen
				DOMTreeWalker* subwalker = createElementWalker(sibling,NULL,false);
				DOMElement* subsibling = (DOMElement*)subwalker->firstChild();
				while(subsibling)
				{
					mergeElements(defptr->getValue(conv(subsibling->getTagName())), subsibling);	
					subsibling = (DOMElement*)subwalker->nextSibling();
				}

				// List mit Komponenten gleichen Typs
				vector<Component*>& rv = config->_components[node->rtti()];
				
				// Index und Scope setzen
				node->setIndex(rv.size());
				node->setScope(sibling);

				// Koordinaten setzen, falls vorhanden
				if(elem->hasAttribute(conv("coords")))
					node->setCoordinates(conv(elem->getAttribute(conv("coords"))));
				
				// Knoten in Konfiguration aufnehmen
				rv.push_back(node);

				//std::cerr << "Create: " << node->toString() << std::endl;
				
				updateConnectables(node, &nodes);
				assert(nodes.getValue(node->getAttribute("id")) == node);
			}

			// Analyser
			else if(equals(sibling->getTagName(), "analyser"))
			{
				vector<XMLElement*>& analysers = measurements[parseAttribute(sibling,"id")];
				
				// Annahme: Knoten noch nicht bearbeitet
				assert(analysers.empty());
				
				std::string value = (std::string)parseAttribute(sibling, "measureVar");
				char* vars = strcpy(new char[value.size()+1], value.c_str()); 	
				char* delims = ";";
				char* expr = strtok(vars, delims);
		
				while(expr) 
				{	
					ValueString vs(expr);

					if(vs.isNumber())
					{
						// Annahme: Analyser aufsteigend sortiert
						assert((uint)vs == 
								(uint)config->_analysers[(uint)vs-1]->getAttribute("measureVar"));
						
						analysers.push_back(config->_analysers[(uint)vs-1]);		
						
						expr = strtok(NULL, delims);
						continue;
					}
					else
						_ERROR("Invalid measue variable (" << (string)vs << ")"
								<< " for analyser " << (string)parseAttribute(sibling, "index") 
								<< "!");
				}

				// Annahme: Mindestens eine Messung
				assert(!analysers.empty());
				assert(!measurements[parseAttribute(sibling,"id")].empty());
				
				delete [] vars;
			}
		}
		
		// Netzkanten
		else if(isConnector(sibling))
		{
			// Leitungen
			if(isDerivedFrom(_schemaInformation, sibling, "WireComponentType"))
			{
				Wire* wire = NULL;
				string key = parseAttribute(sibling, "index");
				
				if(key != _UNDEFINED_)
				{
					wire = (Wire*)_wires.getValue(key);
					
					if(!wire)
					{
						wire = (Wire*)FactoryElement::Factory::self().
							create(conv(sibling->getTagName()));

						// Liste mit Komponenten gleichen Typs
						vector<Component*>& rv = config->_components[wire->rtti()];
						
						// Index und Scope setzen
						wire->setScope(sibling);
						wire->setIndex(rv.size());	
						
						// Leitung in Konfiguration aufnehmen
						rv.push_back(wire);
						
						// Hashmap des Parsers aktualisieren
						_wires[key] = wire;
					}
					
					assert(wire);
					
					if((string)parseAttribute(sibling, "srcType") == "elemIO")
					{
						//assert(!wire->From());
						if(wire->From())
							_ERROR("Corrupted component seen: " << wire->toString())
						
						Node* node = (Node*)nodes.getValue(parseAttribute(sibling, "fromNode"));
						assert(node);
						
						wire->setOutIndex(node->addOutWire(wire));
						wire->setFrom(node);
					}
									
					if((string)parseAttribute(sibling, "destType") == "elemIO")
					{
						//assert(!wire->To());
						if(wire->To())
							_ERROR("Corrupted component seen: " << wire->toString())

						Node* node = (Node*)nodes.getValue(parseAttribute(sibling, "toNode"));
						assert(node);
						
						wire->setInIndex(node->addInWire(wire));
						wire->setTo(node);
					}
				}
			}

			// Messung
			else if(equals(sibling->getTagName(), "obsline"))
			{
				assert(measurements.find(parseAttribute(sibling,"fromNode"))!=measurements.end());
				
				// Messung beginnt an Analyser-Objekt..
				vector<XMLElement*>& analysers = measurements[parseAttribute(sibling, "fromNode")];

				for(uint i=0; i<analysers.size(); i++)
				{
					// ..und endet an Komponente
					Node* node = (Node*)nodes.getValue(parseAttribute(sibling, "toNode"));
					assert(node);
				
					((Analyser*)analysers[i])->push_back(node);
				}
			}
		}
		
		sibling = (DOMElement*)walker->nextSibling();
	}

	// Metakomponenten bearbeiten
	for(unsigned int i = 0; i<metacomponents.size(); i++)
	{
		initConfig(metacomponents[i], defptr?defptr:previous, config);
	}
}
